"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .basesdk import BaseSDK
from apideck_unify import models, utils
from apideck_unify._hooks import HookContext
from apideck_unify.types import OptionalNullable, UNSET
from apideck_unify.utils import get_security_from_env
from apideck_unify.utils.unmarshal_json_response import unmarshal_json_response
from jsonpath import JSONPath
from typing import Any, Dict, List, Mapping, Optional, Union


class Contacts(BaseSDK):
    def list(
        self,
        *,
        raw: Optional[bool] = False,
        consumer_id: Optional[str] = None,
        app_id: Optional[str] = None,
        service_id: Optional[str] = None,
        cursor: OptionalNullable[str] = UNSET,
        limit: Optional[int] = 20,
        filter_: Optional[
            Union[models.ContactsFilter, models.ContactsFilterTypedDict]
        ] = None,
        sort: Optional[Union[models.ContactsSort, models.ContactsSortTypedDict]] = None,
        pass_through: Optional[Dict[str, Any]] = None,
        fields: OptionalNullable[str] = UNSET,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> Optional[models.CrmContactsAllResponse]:
        r"""List contacts

        List contacts

        :param raw: Include raw response. Mostly used for debugging purposes
        :param consumer_id: ID of the consumer which you want to get or push data from
        :param app_id: The ID of your Unify application
        :param service_id: Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API.
        :param cursor: Cursor to start from. You can find cursors for next/previous pages in the meta.cursors property of the response.
        :param limit: Number of results to return. Minimum 1, Maximum 200, Default 20
        :param filter_: Apply filters
        :param sort: Apply sorting
        :param pass_through: Optional unmapped key/values that will be passed through to downstream as query parameters. Ie: ?pass_through[search]=leads becomes ?search=leads
        :param fields: The 'fields' parameter allows API users to specify the fields they want to include in the API response. If this parameter is not present, the API will return all available fields. If this parameter is present, only the fields specified in the comma-separated string will be included in the response. Nested properties can also be requested by using a dot notation. <br /><br />Example: `fields=name,email,addresses.city`<br /><br />In the example above, the response will only include the fields \"name\", \"email\" and \"addresses.city\". If any other fields are available, they will be excluded.
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CrmContactsAllRequest(
            raw=raw,
            consumer_id=consumer_id,
            app_id=app_id,
            service_id=service_id,
            cursor=cursor,
            limit=limit,
            filter_=utils.get_pydantic_model(filter_, Optional[models.ContactsFilter]),
            sort=utils.get_pydantic_model(sort, Optional[models.ContactsSort]),
            pass_through=pass_through,
            fields=fields,
        )

        req = self._build_request(
            method="GET",
            path="/crm/contacts",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            _globals=models.CrmContactsAllGlobals(
                consumer_id=self.sdk_configuration.globals.consumer_id,
                app_id=self.sdk_configuration.globals.app_id,
            ),
            security=self.sdk_configuration.security,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config
            else:
                retries = utils.RetryConfig(
                    "backoff", utils.BackoffStrategy(500, 60000, 1.5, 900000), True
                )

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["408", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="crm.contactsAll",
                oauth2_scopes=None,
                security_source=get_security_from_env(
                    self.sdk_configuration.security, models.Security
                ),
            ),
            request=req,
            error_status_codes=["400", "401", "402", "404", "422", "4XX", "5XX"],
            retry_config=retry_config,
        )

        def next_func() -> Optional[models.CrmContactsAllResponse]:
            body = utils.unmarshal_json(http_res.text, Union[Dict[Any, Any], List[Any]])
            next_cursor = JSONPath("$.meta.cursors.next").parse(body)

            if len(next_cursor) == 0:
                return None

            next_cursor = next_cursor[0]
            if next_cursor is None or str(next_cursor).strip() == "":
                return None

            return self.list(
                raw=raw,
                consumer_id=consumer_id,
                app_id=app_id,
                service_id=service_id,
                cursor=next_cursor,
                limit=limit,
                filter_=filter_,
                sort=sort,
                pass_through=pass_through,
                fields=fields,
                retries=retries,
            )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return models.CrmContactsAllResponse(
                get_contacts_response=unmarshal_json_response(
                    Optional[models.GetContactsResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
                next=next_func,
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.BadRequestResponseData, http_res
            )
            raise models.BadRequestResponse(response_data, http_res)
        if utils.match_response(http_res, "401", "application/json"):
            response_data = unmarshal_json_response(
                models.UnauthorizedResponseData, http_res
            )
            raise models.UnauthorizedResponse(response_data, http_res)
        if utils.match_response(http_res, "402", "application/json"):
            response_data = unmarshal_json_response(
                models.PaymentRequiredResponseData, http_res
            )
            raise models.PaymentRequiredResponse(response_data, http_res)
        if utils.match_response(http_res, "404", "application/json"):
            response_data = unmarshal_json_response(
                models.NotFoundResponseData, http_res
            )
            raise models.NotFoundResponse(response_data, http_res)
        if utils.match_response(http_res, "422", "application/json"):
            response_data = unmarshal_json_response(
                models.UnprocessableResponseData, http_res
            )
            raise models.UnprocessableResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "default", "application/json"):
            return models.CrmContactsAllResponse(
                unexpected_error_response=unmarshal_json_response(
                    Optional[models.UnexpectedErrorResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
                next=next_func,
            )

        raise models.APIError("Unexpected response received", http_res)

    async def list_async(
        self,
        *,
        raw: Optional[bool] = False,
        consumer_id: Optional[str] = None,
        app_id: Optional[str] = None,
        service_id: Optional[str] = None,
        cursor: OptionalNullable[str] = UNSET,
        limit: Optional[int] = 20,
        filter_: Optional[
            Union[models.ContactsFilter, models.ContactsFilterTypedDict]
        ] = None,
        sort: Optional[Union[models.ContactsSort, models.ContactsSortTypedDict]] = None,
        pass_through: Optional[Dict[str, Any]] = None,
        fields: OptionalNullable[str] = UNSET,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> Optional[models.CrmContactsAllResponse]:
        r"""List contacts

        List contacts

        :param raw: Include raw response. Mostly used for debugging purposes
        :param consumer_id: ID of the consumer which you want to get or push data from
        :param app_id: The ID of your Unify application
        :param service_id: Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API.
        :param cursor: Cursor to start from. You can find cursors for next/previous pages in the meta.cursors property of the response.
        :param limit: Number of results to return. Minimum 1, Maximum 200, Default 20
        :param filter_: Apply filters
        :param sort: Apply sorting
        :param pass_through: Optional unmapped key/values that will be passed through to downstream as query parameters. Ie: ?pass_through[search]=leads becomes ?search=leads
        :param fields: The 'fields' parameter allows API users to specify the fields they want to include in the API response. If this parameter is not present, the API will return all available fields. If this parameter is present, only the fields specified in the comma-separated string will be included in the response. Nested properties can also be requested by using a dot notation. <br /><br />Example: `fields=name,email,addresses.city`<br /><br />In the example above, the response will only include the fields \"name\", \"email\" and \"addresses.city\". If any other fields are available, they will be excluded.
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CrmContactsAllRequest(
            raw=raw,
            consumer_id=consumer_id,
            app_id=app_id,
            service_id=service_id,
            cursor=cursor,
            limit=limit,
            filter_=utils.get_pydantic_model(filter_, Optional[models.ContactsFilter]),
            sort=utils.get_pydantic_model(sort, Optional[models.ContactsSort]),
            pass_through=pass_through,
            fields=fields,
        )

        req = self._build_request_async(
            method="GET",
            path="/crm/contacts",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            _globals=models.CrmContactsAllGlobals(
                consumer_id=self.sdk_configuration.globals.consumer_id,
                app_id=self.sdk_configuration.globals.app_id,
            ),
            security=self.sdk_configuration.security,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config
            else:
                retries = utils.RetryConfig(
                    "backoff", utils.BackoffStrategy(500, 60000, 1.5, 900000), True
                )

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["408", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="crm.contactsAll",
                oauth2_scopes=None,
                security_source=get_security_from_env(
                    self.sdk_configuration.security, models.Security
                ),
            ),
            request=req,
            error_status_codes=["400", "401", "402", "404", "422", "4XX", "5XX"],
            retry_config=retry_config,
        )

        def next_func() -> Optional[models.CrmContactsAllResponse]:
            body = utils.unmarshal_json(http_res.text, Union[Dict[Any, Any], List[Any]])
            next_cursor = JSONPath("$.meta.cursors.next").parse(body)

            if len(next_cursor) == 0:
                return None

            next_cursor = next_cursor[0]
            if next_cursor is None or str(next_cursor).strip() == "":
                return None

            return self.list(
                raw=raw,
                consumer_id=consumer_id,
                app_id=app_id,
                service_id=service_id,
                cursor=next_cursor,
                limit=limit,
                filter_=filter_,
                sort=sort,
                pass_through=pass_through,
                fields=fields,
                retries=retries,
            )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return models.CrmContactsAllResponse(
                get_contacts_response=unmarshal_json_response(
                    Optional[models.GetContactsResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
                next=next_func,
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.BadRequestResponseData, http_res
            )
            raise models.BadRequestResponse(response_data, http_res)
        if utils.match_response(http_res, "401", "application/json"):
            response_data = unmarshal_json_response(
                models.UnauthorizedResponseData, http_res
            )
            raise models.UnauthorizedResponse(response_data, http_res)
        if utils.match_response(http_res, "402", "application/json"):
            response_data = unmarshal_json_response(
                models.PaymentRequiredResponseData, http_res
            )
            raise models.PaymentRequiredResponse(response_data, http_res)
        if utils.match_response(http_res, "404", "application/json"):
            response_data = unmarshal_json_response(
                models.NotFoundResponseData, http_res
            )
            raise models.NotFoundResponse(response_data, http_res)
        if utils.match_response(http_res, "422", "application/json"):
            response_data = unmarshal_json_response(
                models.UnprocessableResponseData, http_res
            )
            raise models.UnprocessableResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "default", "application/json"):
            return models.CrmContactsAllResponse(
                unexpected_error_response=unmarshal_json_response(
                    Optional[models.UnexpectedErrorResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
                next=next_func,
            )

        raise models.APIError("Unexpected response received", http_res)

    def create(
        self,
        *,
        raw: Optional[bool] = False,
        consumer_id: Optional[str] = None,
        app_id: Optional[str] = None,
        service_id: Optional[str] = None,
        name: OptionalNullable[str] = UNSET,
        owner_id: OptionalNullable[str] = UNSET,
        type_: OptionalNullable[models.ContactType] = UNSET,
        company_id: OptionalNullable[str] = UNSET,
        company_name: OptionalNullable[str] = UNSET,
        lead_id: OptionalNullable[str] = UNSET,
        first_name: OptionalNullable[str] = UNSET,
        middle_name: OptionalNullable[str] = UNSET,
        last_name: OptionalNullable[str] = UNSET,
        prefix: OptionalNullable[str] = UNSET,
        suffix: OptionalNullable[str] = UNSET,
        title: OptionalNullable[str] = UNSET,
        department: OptionalNullable[str] = UNSET,
        language: OptionalNullable[str] = UNSET,
        gender: OptionalNullable[models.ContactGender] = UNSET,
        birthday: OptionalNullable[str] = UNSET,
        image: OptionalNullable[str] = UNSET,
        photo_url: OptionalNullable[str] = UNSET,
        lead_source: OptionalNullable[str] = UNSET,
        fax: OptionalNullable[str] = UNSET,
        description: OptionalNullable[str] = UNSET,
        current_balance: OptionalNullable[float] = UNSET,
        status: OptionalNullable[str] = UNSET,
        active: OptionalNullable[bool] = UNSET,
        websites: Optional[
            Union[List[models.Website], List[models.WebsiteTypedDict]]
        ] = None,
        addresses: Optional[
            Union[List[models.Address], List[models.AddressTypedDict]]
        ] = None,
        social_links: Optional[
            Union[List[models.SocialLink], List[models.SocialLinkTypedDict]]
        ] = None,
        phone_numbers: Optional[
            Union[List[models.PhoneNumber], List[models.PhoneNumberTypedDict]]
        ] = None,
        emails: Optional[Union[List[models.Email], List[models.EmailTypedDict]]] = None,
        email_domain: OptionalNullable[str] = UNSET,
        custom_fields: OptionalNullable[
            Union[List[models.CustomField], List[models.CustomFieldTypedDict]]
        ] = UNSET,
        tags: OptionalNullable[List[str]] = UNSET,
        opportunity_ids: Optional[List[str]] = None,
        pass_through: Optional[
            Union[List[models.PassThroughBody], List[models.PassThroughBodyTypedDict]]
        ] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.CrmContactsAddResponse:
        r"""Create contact

        Create contact

        :param raw: Include raw response. Mostly used for debugging purposes
        :param consumer_id: ID of the consumer which you want to get or push data from
        :param app_id: The ID of your Unify application
        :param service_id: Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API.
        :param name: Full name of the contact.
        :param owner_id: The owner of the contact.
        :param type: The type of the contact.
        :param company_id: The company the contact is associated with.
        :param company_name: The name of the company the contact is associated with.
        :param lead_id: The lead the contact is associated with.
        :param first_name: The first name of the contact.
        :param middle_name: The middle name of the contact.
        :param last_name: The last name of the contact.
        :param prefix: The prefix of the contact.
        :param suffix: The suffix of the contact.
        :param title: The job title of the contact.
        :param department: The department of the contact.
        :param language: language code according to ISO 639-1. For the United States - EN
        :param gender: The gender of the contact.
        :param birthday: The birthday of the contact.
        :param image:
        :param photo_url: The URL of the photo of a person.
        :param lead_source: The lead source of the contact.
        :param fax: The fax number of the contact.
        :param description: The description of the contact.
        :param current_balance: The current balance of the contact.
        :param status: The status of the contact.
        :param active: The active status of the contact.
        :param websites:
        :param addresses:
        :param social_links:
        :param phone_numbers:
        :param emails:
        :param email_domain:
        :param custom_fields:
        :param tags:
        :param opportunity_ids: The opportunity ids of the contact.
        :param pass_through: The pass_through property allows passing service-specific, custom data or structured modifications in request body when creating or updating resources.
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CrmContactsAddRequest(
            raw=raw,
            consumer_id=consumer_id,
            app_id=app_id,
            service_id=service_id,
            contact=models.ContactInput(
                name=name,
                owner_id=owner_id,
                type=type_,
                company_id=company_id,
                company_name=company_name,
                lead_id=lead_id,
                first_name=first_name,
                middle_name=middle_name,
                last_name=last_name,
                prefix=prefix,
                suffix=suffix,
                title=title,
                department=department,
                language=language,
                gender=gender,
                birthday=birthday,
                image=image,
                photo_url=photo_url,
                lead_source=lead_source,
                fax=fax,
                description=description,
                current_balance=current_balance,
                status=status,
                active=active,
                websites=utils.get_pydantic_model(
                    websites, Optional[List[models.Website]]
                ),
                addresses=utils.get_pydantic_model(
                    addresses, Optional[List[models.Address]]
                ),
                social_links=utils.get_pydantic_model(
                    social_links, Optional[List[models.SocialLink]]
                ),
                phone_numbers=utils.get_pydantic_model(
                    phone_numbers, Optional[List[models.PhoneNumber]]
                ),
                emails=utils.get_pydantic_model(emails, Optional[List[models.Email]]),
                email_domain=email_domain,
                custom_fields=utils.get_pydantic_model(
                    custom_fields, OptionalNullable[List[models.CustomField]]
                ),
                tags=tags,
                opportunity_ids=opportunity_ids,
                pass_through=utils.get_pydantic_model(
                    pass_through, Optional[List[models.PassThroughBody]]
                ),
            ),
        )

        req = self._build_request(
            method="POST",
            path="/crm/contacts",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            _globals=models.CrmContactsAddGlobals(
                consumer_id=self.sdk_configuration.globals.consumer_id,
                app_id=self.sdk_configuration.globals.app_id,
            ),
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request.contact, False, False, "json", models.ContactInput
            ),
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config
            else:
                retries = utils.RetryConfig(
                    "backoff", utils.BackoffStrategy(500, 60000, 1.5, 900000), True
                )

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["408", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="crm.contactsAdd",
                oauth2_scopes=None,
                security_source=get_security_from_env(
                    self.sdk_configuration.security, models.Security
                ),
            ),
            request=req,
            error_status_codes=["400", "401", "402", "404", "422", "4XX", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "201", "application/json"):
            return models.CrmContactsAddResponse(
                create_contact_response=unmarshal_json_response(
                    Optional[models.CreateContactResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.BadRequestResponseData, http_res
            )
            raise models.BadRequestResponse(response_data, http_res)
        if utils.match_response(http_res, "401", "application/json"):
            response_data = unmarshal_json_response(
                models.UnauthorizedResponseData, http_res
            )
            raise models.UnauthorizedResponse(response_data, http_res)
        if utils.match_response(http_res, "402", "application/json"):
            response_data = unmarshal_json_response(
                models.PaymentRequiredResponseData, http_res
            )
            raise models.PaymentRequiredResponse(response_data, http_res)
        if utils.match_response(http_res, "404", "application/json"):
            response_data = unmarshal_json_response(
                models.NotFoundResponseData, http_res
            )
            raise models.NotFoundResponse(response_data, http_res)
        if utils.match_response(http_res, "422", "application/json"):
            response_data = unmarshal_json_response(
                models.UnprocessableResponseData, http_res
            )
            raise models.UnprocessableResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "default", "application/json"):
            return models.CrmContactsAddResponse(
                unexpected_error_response=unmarshal_json_response(
                    Optional[models.UnexpectedErrorResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )

        raise models.APIError("Unexpected response received", http_res)

    async def create_async(
        self,
        *,
        raw: Optional[bool] = False,
        consumer_id: Optional[str] = None,
        app_id: Optional[str] = None,
        service_id: Optional[str] = None,
        name: OptionalNullable[str] = UNSET,
        owner_id: OptionalNullable[str] = UNSET,
        type_: OptionalNullable[models.ContactType] = UNSET,
        company_id: OptionalNullable[str] = UNSET,
        company_name: OptionalNullable[str] = UNSET,
        lead_id: OptionalNullable[str] = UNSET,
        first_name: OptionalNullable[str] = UNSET,
        middle_name: OptionalNullable[str] = UNSET,
        last_name: OptionalNullable[str] = UNSET,
        prefix: OptionalNullable[str] = UNSET,
        suffix: OptionalNullable[str] = UNSET,
        title: OptionalNullable[str] = UNSET,
        department: OptionalNullable[str] = UNSET,
        language: OptionalNullable[str] = UNSET,
        gender: OptionalNullable[models.ContactGender] = UNSET,
        birthday: OptionalNullable[str] = UNSET,
        image: OptionalNullable[str] = UNSET,
        photo_url: OptionalNullable[str] = UNSET,
        lead_source: OptionalNullable[str] = UNSET,
        fax: OptionalNullable[str] = UNSET,
        description: OptionalNullable[str] = UNSET,
        current_balance: OptionalNullable[float] = UNSET,
        status: OptionalNullable[str] = UNSET,
        active: OptionalNullable[bool] = UNSET,
        websites: Optional[
            Union[List[models.Website], List[models.WebsiteTypedDict]]
        ] = None,
        addresses: Optional[
            Union[List[models.Address], List[models.AddressTypedDict]]
        ] = None,
        social_links: Optional[
            Union[List[models.SocialLink], List[models.SocialLinkTypedDict]]
        ] = None,
        phone_numbers: Optional[
            Union[List[models.PhoneNumber], List[models.PhoneNumberTypedDict]]
        ] = None,
        emails: Optional[Union[List[models.Email], List[models.EmailTypedDict]]] = None,
        email_domain: OptionalNullable[str] = UNSET,
        custom_fields: OptionalNullable[
            Union[List[models.CustomField], List[models.CustomFieldTypedDict]]
        ] = UNSET,
        tags: OptionalNullable[List[str]] = UNSET,
        opportunity_ids: Optional[List[str]] = None,
        pass_through: Optional[
            Union[List[models.PassThroughBody], List[models.PassThroughBodyTypedDict]]
        ] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.CrmContactsAddResponse:
        r"""Create contact

        Create contact

        :param raw: Include raw response. Mostly used for debugging purposes
        :param consumer_id: ID of the consumer which you want to get or push data from
        :param app_id: The ID of your Unify application
        :param service_id: Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API.
        :param name: Full name of the contact.
        :param owner_id: The owner of the contact.
        :param type: The type of the contact.
        :param company_id: The company the contact is associated with.
        :param company_name: The name of the company the contact is associated with.
        :param lead_id: The lead the contact is associated with.
        :param first_name: The first name of the contact.
        :param middle_name: The middle name of the contact.
        :param last_name: The last name of the contact.
        :param prefix: The prefix of the contact.
        :param suffix: The suffix of the contact.
        :param title: The job title of the contact.
        :param department: The department of the contact.
        :param language: language code according to ISO 639-1. For the United States - EN
        :param gender: The gender of the contact.
        :param birthday: The birthday of the contact.
        :param image:
        :param photo_url: The URL of the photo of a person.
        :param lead_source: The lead source of the contact.
        :param fax: The fax number of the contact.
        :param description: The description of the contact.
        :param current_balance: The current balance of the contact.
        :param status: The status of the contact.
        :param active: The active status of the contact.
        :param websites:
        :param addresses:
        :param social_links:
        :param phone_numbers:
        :param emails:
        :param email_domain:
        :param custom_fields:
        :param tags:
        :param opportunity_ids: The opportunity ids of the contact.
        :param pass_through: The pass_through property allows passing service-specific, custom data or structured modifications in request body when creating or updating resources.
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CrmContactsAddRequest(
            raw=raw,
            consumer_id=consumer_id,
            app_id=app_id,
            service_id=service_id,
            contact=models.ContactInput(
                name=name,
                owner_id=owner_id,
                type=type_,
                company_id=company_id,
                company_name=company_name,
                lead_id=lead_id,
                first_name=first_name,
                middle_name=middle_name,
                last_name=last_name,
                prefix=prefix,
                suffix=suffix,
                title=title,
                department=department,
                language=language,
                gender=gender,
                birthday=birthday,
                image=image,
                photo_url=photo_url,
                lead_source=lead_source,
                fax=fax,
                description=description,
                current_balance=current_balance,
                status=status,
                active=active,
                websites=utils.get_pydantic_model(
                    websites, Optional[List[models.Website]]
                ),
                addresses=utils.get_pydantic_model(
                    addresses, Optional[List[models.Address]]
                ),
                social_links=utils.get_pydantic_model(
                    social_links, Optional[List[models.SocialLink]]
                ),
                phone_numbers=utils.get_pydantic_model(
                    phone_numbers, Optional[List[models.PhoneNumber]]
                ),
                emails=utils.get_pydantic_model(emails, Optional[List[models.Email]]),
                email_domain=email_domain,
                custom_fields=utils.get_pydantic_model(
                    custom_fields, OptionalNullable[List[models.CustomField]]
                ),
                tags=tags,
                opportunity_ids=opportunity_ids,
                pass_through=utils.get_pydantic_model(
                    pass_through, Optional[List[models.PassThroughBody]]
                ),
            ),
        )

        req = self._build_request_async(
            method="POST",
            path="/crm/contacts",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            _globals=models.CrmContactsAddGlobals(
                consumer_id=self.sdk_configuration.globals.consumer_id,
                app_id=self.sdk_configuration.globals.app_id,
            ),
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request.contact, False, False, "json", models.ContactInput
            ),
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config
            else:
                retries = utils.RetryConfig(
                    "backoff", utils.BackoffStrategy(500, 60000, 1.5, 900000), True
                )

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["408", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="crm.contactsAdd",
                oauth2_scopes=None,
                security_source=get_security_from_env(
                    self.sdk_configuration.security, models.Security
                ),
            ),
            request=req,
            error_status_codes=["400", "401", "402", "404", "422", "4XX", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "201", "application/json"):
            return models.CrmContactsAddResponse(
                create_contact_response=unmarshal_json_response(
                    Optional[models.CreateContactResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.BadRequestResponseData, http_res
            )
            raise models.BadRequestResponse(response_data, http_res)
        if utils.match_response(http_res, "401", "application/json"):
            response_data = unmarshal_json_response(
                models.UnauthorizedResponseData, http_res
            )
            raise models.UnauthorizedResponse(response_data, http_res)
        if utils.match_response(http_res, "402", "application/json"):
            response_data = unmarshal_json_response(
                models.PaymentRequiredResponseData, http_res
            )
            raise models.PaymentRequiredResponse(response_data, http_res)
        if utils.match_response(http_res, "404", "application/json"):
            response_data = unmarshal_json_response(
                models.NotFoundResponseData, http_res
            )
            raise models.NotFoundResponse(response_data, http_res)
        if utils.match_response(http_res, "422", "application/json"):
            response_data = unmarshal_json_response(
                models.UnprocessableResponseData, http_res
            )
            raise models.UnprocessableResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "default", "application/json"):
            return models.CrmContactsAddResponse(
                unexpected_error_response=unmarshal_json_response(
                    Optional[models.UnexpectedErrorResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )

        raise models.APIError("Unexpected response received", http_res)

    def get(
        self,
        *,
        id: str,
        consumer_id: Optional[str] = None,
        app_id: Optional[str] = None,
        service_id: Optional[str] = None,
        raw: Optional[bool] = False,
        fields: OptionalNullable[str] = UNSET,
        filter_: Optional[
            Union[models.ContactsFilter, models.ContactsFilterTypedDict]
        ] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.CrmContactsOneResponse:
        r"""Get contact

        Get contact

        :param id: ID of the record you are acting upon.
        :param consumer_id: ID of the consumer which you want to get or push data from
        :param app_id: The ID of your Unify application
        :param service_id: Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API.
        :param raw: Include raw response. Mostly used for debugging purposes
        :param fields: The 'fields' parameter allows API users to specify the fields they want to include in the API response. If this parameter is not present, the API will return all available fields. If this parameter is present, only the fields specified in the comma-separated string will be included in the response. Nested properties can also be requested by using a dot notation. <br /><br />Example: `fields=name,email,addresses.city`<br /><br />In the example above, the response will only include the fields \"name\", \"email\" and \"addresses.city\". If any other fields are available, they will be excluded.
        :param filter_: Apply filters
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CrmContactsOneRequest(
            id=id,
            consumer_id=consumer_id,
            app_id=app_id,
            service_id=service_id,
            raw=raw,
            fields=fields,
            filter_=utils.get_pydantic_model(filter_, Optional[models.ContactsFilter]),
        )

        req = self._build_request(
            method="GET",
            path="/crm/contacts/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            _globals=models.CrmContactsOneGlobals(
                consumer_id=self.sdk_configuration.globals.consumer_id,
                app_id=self.sdk_configuration.globals.app_id,
            ),
            security=self.sdk_configuration.security,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config
            else:
                retries = utils.RetryConfig(
                    "backoff", utils.BackoffStrategy(500, 60000, 1.5, 900000), True
                )

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["408", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="crm.contactsOne",
                oauth2_scopes=None,
                security_source=get_security_from_env(
                    self.sdk_configuration.security, models.Security
                ),
            ),
            request=req,
            error_status_codes=["400", "401", "402", "404", "422", "4XX", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return models.CrmContactsOneResponse(
                get_contact_response=unmarshal_json_response(
                    Optional[models.GetContactResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.BadRequestResponseData, http_res
            )
            raise models.BadRequestResponse(response_data, http_res)
        if utils.match_response(http_res, "401", "application/json"):
            response_data = unmarshal_json_response(
                models.UnauthorizedResponseData, http_res
            )
            raise models.UnauthorizedResponse(response_data, http_res)
        if utils.match_response(http_res, "402", "application/json"):
            response_data = unmarshal_json_response(
                models.PaymentRequiredResponseData, http_res
            )
            raise models.PaymentRequiredResponse(response_data, http_res)
        if utils.match_response(http_res, "404", "application/json"):
            response_data = unmarshal_json_response(
                models.NotFoundResponseData, http_res
            )
            raise models.NotFoundResponse(response_data, http_res)
        if utils.match_response(http_res, "422", "application/json"):
            response_data = unmarshal_json_response(
                models.UnprocessableResponseData, http_res
            )
            raise models.UnprocessableResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "default", "application/json"):
            return models.CrmContactsOneResponse(
                unexpected_error_response=unmarshal_json_response(
                    Optional[models.UnexpectedErrorResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )

        raise models.APIError("Unexpected response received", http_res)

    async def get_async(
        self,
        *,
        id: str,
        consumer_id: Optional[str] = None,
        app_id: Optional[str] = None,
        service_id: Optional[str] = None,
        raw: Optional[bool] = False,
        fields: OptionalNullable[str] = UNSET,
        filter_: Optional[
            Union[models.ContactsFilter, models.ContactsFilterTypedDict]
        ] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.CrmContactsOneResponse:
        r"""Get contact

        Get contact

        :param id: ID of the record you are acting upon.
        :param consumer_id: ID of the consumer which you want to get or push data from
        :param app_id: The ID of your Unify application
        :param service_id: Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API.
        :param raw: Include raw response. Mostly used for debugging purposes
        :param fields: The 'fields' parameter allows API users to specify the fields they want to include in the API response. If this parameter is not present, the API will return all available fields. If this parameter is present, only the fields specified in the comma-separated string will be included in the response. Nested properties can also be requested by using a dot notation. <br /><br />Example: `fields=name,email,addresses.city`<br /><br />In the example above, the response will only include the fields \"name\", \"email\" and \"addresses.city\". If any other fields are available, they will be excluded.
        :param filter_: Apply filters
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CrmContactsOneRequest(
            id=id,
            consumer_id=consumer_id,
            app_id=app_id,
            service_id=service_id,
            raw=raw,
            fields=fields,
            filter_=utils.get_pydantic_model(filter_, Optional[models.ContactsFilter]),
        )

        req = self._build_request_async(
            method="GET",
            path="/crm/contacts/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            _globals=models.CrmContactsOneGlobals(
                consumer_id=self.sdk_configuration.globals.consumer_id,
                app_id=self.sdk_configuration.globals.app_id,
            ),
            security=self.sdk_configuration.security,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config
            else:
                retries = utils.RetryConfig(
                    "backoff", utils.BackoffStrategy(500, 60000, 1.5, 900000), True
                )

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["408", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="crm.contactsOne",
                oauth2_scopes=None,
                security_source=get_security_from_env(
                    self.sdk_configuration.security, models.Security
                ),
            ),
            request=req,
            error_status_codes=["400", "401", "402", "404", "422", "4XX", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return models.CrmContactsOneResponse(
                get_contact_response=unmarshal_json_response(
                    Optional[models.GetContactResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.BadRequestResponseData, http_res
            )
            raise models.BadRequestResponse(response_data, http_res)
        if utils.match_response(http_res, "401", "application/json"):
            response_data = unmarshal_json_response(
                models.UnauthorizedResponseData, http_res
            )
            raise models.UnauthorizedResponse(response_data, http_res)
        if utils.match_response(http_res, "402", "application/json"):
            response_data = unmarshal_json_response(
                models.PaymentRequiredResponseData, http_res
            )
            raise models.PaymentRequiredResponse(response_data, http_res)
        if utils.match_response(http_res, "404", "application/json"):
            response_data = unmarshal_json_response(
                models.NotFoundResponseData, http_res
            )
            raise models.NotFoundResponse(response_data, http_res)
        if utils.match_response(http_res, "422", "application/json"):
            response_data = unmarshal_json_response(
                models.UnprocessableResponseData, http_res
            )
            raise models.UnprocessableResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "default", "application/json"):
            return models.CrmContactsOneResponse(
                unexpected_error_response=unmarshal_json_response(
                    Optional[models.UnexpectedErrorResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )

        raise models.APIError("Unexpected response received", http_res)

    def update(
        self,
        *,
        id: str,
        consumer_id: Optional[str] = None,
        app_id: Optional[str] = None,
        service_id: Optional[str] = None,
        raw: Optional[bool] = False,
        name: OptionalNullable[str] = UNSET,
        owner_id: OptionalNullable[str] = UNSET,
        type_: OptionalNullable[models.ContactType] = UNSET,
        company_id: OptionalNullable[str] = UNSET,
        company_name: OptionalNullable[str] = UNSET,
        lead_id: OptionalNullable[str] = UNSET,
        first_name: OptionalNullable[str] = UNSET,
        middle_name: OptionalNullable[str] = UNSET,
        last_name: OptionalNullable[str] = UNSET,
        prefix: OptionalNullable[str] = UNSET,
        suffix: OptionalNullable[str] = UNSET,
        title: OptionalNullable[str] = UNSET,
        department: OptionalNullable[str] = UNSET,
        language: OptionalNullable[str] = UNSET,
        gender: OptionalNullable[models.ContactGender] = UNSET,
        birthday: OptionalNullable[str] = UNSET,
        image: OptionalNullable[str] = UNSET,
        photo_url: OptionalNullable[str] = UNSET,
        lead_source: OptionalNullable[str] = UNSET,
        fax: OptionalNullable[str] = UNSET,
        description: OptionalNullable[str] = UNSET,
        current_balance: OptionalNullable[float] = UNSET,
        status: OptionalNullable[str] = UNSET,
        active: OptionalNullable[bool] = UNSET,
        websites: Optional[
            Union[List[models.Website], List[models.WebsiteTypedDict]]
        ] = None,
        addresses: Optional[
            Union[List[models.Address], List[models.AddressTypedDict]]
        ] = None,
        social_links: Optional[
            Union[List[models.SocialLink], List[models.SocialLinkTypedDict]]
        ] = None,
        phone_numbers: Optional[
            Union[List[models.PhoneNumber], List[models.PhoneNumberTypedDict]]
        ] = None,
        emails: Optional[Union[List[models.Email], List[models.EmailTypedDict]]] = None,
        email_domain: OptionalNullable[str] = UNSET,
        custom_fields: OptionalNullable[
            Union[List[models.CustomField], List[models.CustomFieldTypedDict]]
        ] = UNSET,
        tags: OptionalNullable[List[str]] = UNSET,
        opportunity_ids: Optional[List[str]] = None,
        pass_through: Optional[
            Union[List[models.PassThroughBody], List[models.PassThroughBodyTypedDict]]
        ] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.CrmContactsUpdateResponse:
        r"""Update contact

        Update contact

        :param id: ID of the record you are acting upon.
        :param consumer_id: ID of the consumer which you want to get or push data from
        :param app_id: The ID of your Unify application
        :param service_id: Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API.
        :param raw: Include raw response. Mostly used for debugging purposes
        :param name: Full name of the contact.
        :param owner_id: The owner of the contact.
        :param type: The type of the contact.
        :param company_id: The company the contact is associated with.
        :param company_name: The name of the company the contact is associated with.
        :param lead_id: The lead the contact is associated with.
        :param first_name: The first name of the contact.
        :param middle_name: The middle name of the contact.
        :param last_name: The last name of the contact.
        :param prefix: The prefix of the contact.
        :param suffix: The suffix of the contact.
        :param title: The job title of the contact.
        :param department: The department of the contact.
        :param language: language code according to ISO 639-1. For the United States - EN
        :param gender: The gender of the contact.
        :param birthday: The birthday of the contact.
        :param image:
        :param photo_url: The URL of the photo of a person.
        :param lead_source: The lead source of the contact.
        :param fax: The fax number of the contact.
        :param description: The description of the contact.
        :param current_balance: The current balance of the contact.
        :param status: The status of the contact.
        :param active: The active status of the contact.
        :param websites:
        :param addresses:
        :param social_links:
        :param phone_numbers:
        :param emails:
        :param email_domain:
        :param custom_fields:
        :param tags:
        :param opportunity_ids: The opportunity ids of the contact.
        :param pass_through: The pass_through property allows passing service-specific, custom data or structured modifications in request body when creating or updating resources.
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CrmContactsUpdateRequest(
            id=id,
            consumer_id=consumer_id,
            app_id=app_id,
            service_id=service_id,
            raw=raw,
            contact=models.ContactInput(
                name=name,
                owner_id=owner_id,
                type=type_,
                company_id=company_id,
                company_name=company_name,
                lead_id=lead_id,
                first_name=first_name,
                middle_name=middle_name,
                last_name=last_name,
                prefix=prefix,
                suffix=suffix,
                title=title,
                department=department,
                language=language,
                gender=gender,
                birthday=birthday,
                image=image,
                photo_url=photo_url,
                lead_source=lead_source,
                fax=fax,
                description=description,
                current_balance=current_balance,
                status=status,
                active=active,
                websites=utils.get_pydantic_model(
                    websites, Optional[List[models.Website]]
                ),
                addresses=utils.get_pydantic_model(
                    addresses, Optional[List[models.Address]]
                ),
                social_links=utils.get_pydantic_model(
                    social_links, Optional[List[models.SocialLink]]
                ),
                phone_numbers=utils.get_pydantic_model(
                    phone_numbers, Optional[List[models.PhoneNumber]]
                ),
                emails=utils.get_pydantic_model(emails, Optional[List[models.Email]]),
                email_domain=email_domain,
                custom_fields=utils.get_pydantic_model(
                    custom_fields, OptionalNullable[List[models.CustomField]]
                ),
                tags=tags,
                opportunity_ids=opportunity_ids,
                pass_through=utils.get_pydantic_model(
                    pass_through, Optional[List[models.PassThroughBody]]
                ),
            ),
        )

        req = self._build_request(
            method="PATCH",
            path="/crm/contacts/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            _globals=models.CrmContactsUpdateGlobals(
                consumer_id=self.sdk_configuration.globals.consumer_id,
                app_id=self.sdk_configuration.globals.app_id,
            ),
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request.contact, False, False, "json", models.ContactInput
            ),
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config
            else:
                retries = utils.RetryConfig(
                    "backoff", utils.BackoffStrategy(500, 60000, 1.5, 900000), True
                )

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["408", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="crm.contactsUpdate",
                oauth2_scopes=None,
                security_source=get_security_from_env(
                    self.sdk_configuration.security, models.Security
                ),
            ),
            request=req,
            error_status_codes=["400", "401", "402", "404", "422", "4XX", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return models.CrmContactsUpdateResponse(
                update_contact_response=unmarshal_json_response(
                    Optional[models.UpdateContactResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.BadRequestResponseData, http_res
            )
            raise models.BadRequestResponse(response_data, http_res)
        if utils.match_response(http_res, "401", "application/json"):
            response_data = unmarshal_json_response(
                models.UnauthorizedResponseData, http_res
            )
            raise models.UnauthorizedResponse(response_data, http_res)
        if utils.match_response(http_res, "402", "application/json"):
            response_data = unmarshal_json_response(
                models.PaymentRequiredResponseData, http_res
            )
            raise models.PaymentRequiredResponse(response_data, http_res)
        if utils.match_response(http_res, "404", "application/json"):
            response_data = unmarshal_json_response(
                models.NotFoundResponseData, http_res
            )
            raise models.NotFoundResponse(response_data, http_res)
        if utils.match_response(http_res, "422", "application/json"):
            response_data = unmarshal_json_response(
                models.UnprocessableResponseData, http_res
            )
            raise models.UnprocessableResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "default", "application/json"):
            return models.CrmContactsUpdateResponse(
                unexpected_error_response=unmarshal_json_response(
                    Optional[models.UnexpectedErrorResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )

        raise models.APIError("Unexpected response received", http_res)

    async def update_async(
        self,
        *,
        id: str,
        consumer_id: Optional[str] = None,
        app_id: Optional[str] = None,
        service_id: Optional[str] = None,
        raw: Optional[bool] = False,
        name: OptionalNullable[str] = UNSET,
        owner_id: OptionalNullable[str] = UNSET,
        type_: OptionalNullable[models.ContactType] = UNSET,
        company_id: OptionalNullable[str] = UNSET,
        company_name: OptionalNullable[str] = UNSET,
        lead_id: OptionalNullable[str] = UNSET,
        first_name: OptionalNullable[str] = UNSET,
        middle_name: OptionalNullable[str] = UNSET,
        last_name: OptionalNullable[str] = UNSET,
        prefix: OptionalNullable[str] = UNSET,
        suffix: OptionalNullable[str] = UNSET,
        title: OptionalNullable[str] = UNSET,
        department: OptionalNullable[str] = UNSET,
        language: OptionalNullable[str] = UNSET,
        gender: OptionalNullable[models.ContactGender] = UNSET,
        birthday: OptionalNullable[str] = UNSET,
        image: OptionalNullable[str] = UNSET,
        photo_url: OptionalNullable[str] = UNSET,
        lead_source: OptionalNullable[str] = UNSET,
        fax: OptionalNullable[str] = UNSET,
        description: OptionalNullable[str] = UNSET,
        current_balance: OptionalNullable[float] = UNSET,
        status: OptionalNullable[str] = UNSET,
        active: OptionalNullable[bool] = UNSET,
        websites: Optional[
            Union[List[models.Website], List[models.WebsiteTypedDict]]
        ] = None,
        addresses: Optional[
            Union[List[models.Address], List[models.AddressTypedDict]]
        ] = None,
        social_links: Optional[
            Union[List[models.SocialLink], List[models.SocialLinkTypedDict]]
        ] = None,
        phone_numbers: Optional[
            Union[List[models.PhoneNumber], List[models.PhoneNumberTypedDict]]
        ] = None,
        emails: Optional[Union[List[models.Email], List[models.EmailTypedDict]]] = None,
        email_domain: OptionalNullable[str] = UNSET,
        custom_fields: OptionalNullable[
            Union[List[models.CustomField], List[models.CustomFieldTypedDict]]
        ] = UNSET,
        tags: OptionalNullable[List[str]] = UNSET,
        opportunity_ids: Optional[List[str]] = None,
        pass_through: Optional[
            Union[List[models.PassThroughBody], List[models.PassThroughBodyTypedDict]]
        ] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.CrmContactsUpdateResponse:
        r"""Update contact

        Update contact

        :param id: ID of the record you are acting upon.
        :param consumer_id: ID of the consumer which you want to get or push data from
        :param app_id: The ID of your Unify application
        :param service_id: Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API.
        :param raw: Include raw response. Mostly used for debugging purposes
        :param name: Full name of the contact.
        :param owner_id: The owner of the contact.
        :param type: The type of the contact.
        :param company_id: The company the contact is associated with.
        :param company_name: The name of the company the contact is associated with.
        :param lead_id: The lead the contact is associated with.
        :param first_name: The first name of the contact.
        :param middle_name: The middle name of the contact.
        :param last_name: The last name of the contact.
        :param prefix: The prefix of the contact.
        :param suffix: The suffix of the contact.
        :param title: The job title of the contact.
        :param department: The department of the contact.
        :param language: language code according to ISO 639-1. For the United States - EN
        :param gender: The gender of the contact.
        :param birthday: The birthday of the contact.
        :param image:
        :param photo_url: The URL of the photo of a person.
        :param lead_source: The lead source of the contact.
        :param fax: The fax number of the contact.
        :param description: The description of the contact.
        :param current_balance: The current balance of the contact.
        :param status: The status of the contact.
        :param active: The active status of the contact.
        :param websites:
        :param addresses:
        :param social_links:
        :param phone_numbers:
        :param emails:
        :param email_domain:
        :param custom_fields:
        :param tags:
        :param opportunity_ids: The opportunity ids of the contact.
        :param pass_through: The pass_through property allows passing service-specific, custom data or structured modifications in request body when creating or updating resources.
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CrmContactsUpdateRequest(
            id=id,
            consumer_id=consumer_id,
            app_id=app_id,
            service_id=service_id,
            raw=raw,
            contact=models.ContactInput(
                name=name,
                owner_id=owner_id,
                type=type_,
                company_id=company_id,
                company_name=company_name,
                lead_id=lead_id,
                first_name=first_name,
                middle_name=middle_name,
                last_name=last_name,
                prefix=prefix,
                suffix=suffix,
                title=title,
                department=department,
                language=language,
                gender=gender,
                birthday=birthday,
                image=image,
                photo_url=photo_url,
                lead_source=lead_source,
                fax=fax,
                description=description,
                current_balance=current_balance,
                status=status,
                active=active,
                websites=utils.get_pydantic_model(
                    websites, Optional[List[models.Website]]
                ),
                addresses=utils.get_pydantic_model(
                    addresses, Optional[List[models.Address]]
                ),
                social_links=utils.get_pydantic_model(
                    social_links, Optional[List[models.SocialLink]]
                ),
                phone_numbers=utils.get_pydantic_model(
                    phone_numbers, Optional[List[models.PhoneNumber]]
                ),
                emails=utils.get_pydantic_model(emails, Optional[List[models.Email]]),
                email_domain=email_domain,
                custom_fields=utils.get_pydantic_model(
                    custom_fields, OptionalNullable[List[models.CustomField]]
                ),
                tags=tags,
                opportunity_ids=opportunity_ids,
                pass_through=utils.get_pydantic_model(
                    pass_through, Optional[List[models.PassThroughBody]]
                ),
            ),
        )

        req = self._build_request_async(
            method="PATCH",
            path="/crm/contacts/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            _globals=models.CrmContactsUpdateGlobals(
                consumer_id=self.sdk_configuration.globals.consumer_id,
                app_id=self.sdk_configuration.globals.app_id,
            ),
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request.contact, False, False, "json", models.ContactInput
            ),
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config
            else:
                retries = utils.RetryConfig(
                    "backoff", utils.BackoffStrategy(500, 60000, 1.5, 900000), True
                )

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["408", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="crm.contactsUpdate",
                oauth2_scopes=None,
                security_source=get_security_from_env(
                    self.sdk_configuration.security, models.Security
                ),
            ),
            request=req,
            error_status_codes=["400", "401", "402", "404", "422", "4XX", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return models.CrmContactsUpdateResponse(
                update_contact_response=unmarshal_json_response(
                    Optional[models.UpdateContactResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.BadRequestResponseData, http_res
            )
            raise models.BadRequestResponse(response_data, http_res)
        if utils.match_response(http_res, "401", "application/json"):
            response_data = unmarshal_json_response(
                models.UnauthorizedResponseData, http_res
            )
            raise models.UnauthorizedResponse(response_data, http_res)
        if utils.match_response(http_res, "402", "application/json"):
            response_data = unmarshal_json_response(
                models.PaymentRequiredResponseData, http_res
            )
            raise models.PaymentRequiredResponse(response_data, http_res)
        if utils.match_response(http_res, "404", "application/json"):
            response_data = unmarshal_json_response(
                models.NotFoundResponseData, http_res
            )
            raise models.NotFoundResponse(response_data, http_res)
        if utils.match_response(http_res, "422", "application/json"):
            response_data = unmarshal_json_response(
                models.UnprocessableResponseData, http_res
            )
            raise models.UnprocessableResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "default", "application/json"):
            return models.CrmContactsUpdateResponse(
                unexpected_error_response=unmarshal_json_response(
                    Optional[models.UnexpectedErrorResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )

        raise models.APIError("Unexpected response received", http_res)

    def delete(
        self,
        *,
        id: str,
        consumer_id: Optional[str] = None,
        app_id: Optional[str] = None,
        service_id: Optional[str] = None,
        raw: Optional[bool] = False,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.CrmContactsDeleteResponse:
        r"""Delete contact

        Delete contact

        :param id: ID of the record you are acting upon.
        :param consumer_id: ID of the consumer which you want to get or push data from
        :param app_id: The ID of your Unify application
        :param service_id: Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API.
        :param raw: Include raw response. Mostly used for debugging purposes
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CrmContactsDeleteRequest(
            id=id,
            consumer_id=consumer_id,
            app_id=app_id,
            service_id=service_id,
            raw=raw,
        )

        req = self._build_request(
            method="DELETE",
            path="/crm/contacts/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            _globals=models.CrmContactsDeleteGlobals(
                consumer_id=self.sdk_configuration.globals.consumer_id,
                app_id=self.sdk_configuration.globals.app_id,
            ),
            security=self.sdk_configuration.security,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config
            else:
                retries = utils.RetryConfig(
                    "backoff", utils.BackoffStrategy(500, 60000, 1.5, 900000), True
                )

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["408", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="crm.contactsDelete",
                oauth2_scopes=None,
                security_source=get_security_from_env(
                    self.sdk_configuration.security, models.Security
                ),
            ),
            request=req,
            error_status_codes=["400", "401", "402", "404", "422", "4XX", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return models.CrmContactsDeleteResponse(
                delete_contact_response=unmarshal_json_response(
                    Optional[models.DeleteContactResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.BadRequestResponseData, http_res
            )
            raise models.BadRequestResponse(response_data, http_res)
        if utils.match_response(http_res, "401", "application/json"):
            response_data = unmarshal_json_response(
                models.UnauthorizedResponseData, http_res
            )
            raise models.UnauthorizedResponse(response_data, http_res)
        if utils.match_response(http_res, "402", "application/json"):
            response_data = unmarshal_json_response(
                models.PaymentRequiredResponseData, http_res
            )
            raise models.PaymentRequiredResponse(response_data, http_res)
        if utils.match_response(http_res, "404", "application/json"):
            response_data = unmarshal_json_response(
                models.NotFoundResponseData, http_res
            )
            raise models.NotFoundResponse(response_data, http_res)
        if utils.match_response(http_res, "422", "application/json"):
            response_data = unmarshal_json_response(
                models.UnprocessableResponseData, http_res
            )
            raise models.UnprocessableResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "default", "application/json"):
            return models.CrmContactsDeleteResponse(
                unexpected_error_response=unmarshal_json_response(
                    Optional[models.UnexpectedErrorResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )

        raise models.APIError("Unexpected response received", http_res)

    async def delete_async(
        self,
        *,
        id: str,
        consumer_id: Optional[str] = None,
        app_id: Optional[str] = None,
        service_id: Optional[str] = None,
        raw: Optional[bool] = False,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.CrmContactsDeleteResponse:
        r"""Delete contact

        Delete contact

        :param id: ID of the record you are acting upon.
        :param consumer_id: ID of the consumer which you want to get or push data from
        :param app_id: The ID of your Unify application
        :param service_id: Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API.
        :param raw: Include raw response. Mostly used for debugging purposes
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CrmContactsDeleteRequest(
            id=id,
            consumer_id=consumer_id,
            app_id=app_id,
            service_id=service_id,
            raw=raw,
        )

        req = self._build_request_async(
            method="DELETE",
            path="/crm/contacts/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            _globals=models.CrmContactsDeleteGlobals(
                consumer_id=self.sdk_configuration.globals.consumer_id,
                app_id=self.sdk_configuration.globals.app_id,
            ),
            security=self.sdk_configuration.security,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config
            else:
                retries = utils.RetryConfig(
                    "backoff", utils.BackoffStrategy(500, 60000, 1.5, 900000), True
                )

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["408", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="crm.contactsDelete",
                oauth2_scopes=None,
                security_source=get_security_from_env(
                    self.sdk_configuration.security, models.Security
                ),
            ),
            request=req,
            error_status_codes=["400", "401", "402", "404", "422", "4XX", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return models.CrmContactsDeleteResponse(
                delete_contact_response=unmarshal_json_response(
                    Optional[models.DeleteContactResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.BadRequestResponseData, http_res
            )
            raise models.BadRequestResponse(response_data, http_res)
        if utils.match_response(http_res, "401", "application/json"):
            response_data = unmarshal_json_response(
                models.UnauthorizedResponseData, http_res
            )
            raise models.UnauthorizedResponse(response_data, http_res)
        if utils.match_response(http_res, "402", "application/json"):
            response_data = unmarshal_json_response(
                models.PaymentRequiredResponseData, http_res
            )
            raise models.PaymentRequiredResponse(response_data, http_res)
        if utils.match_response(http_res, "404", "application/json"):
            response_data = unmarshal_json_response(
                models.NotFoundResponseData, http_res
            )
            raise models.NotFoundResponse(response_data, http_res)
        if utils.match_response(http_res, "422", "application/json"):
            response_data = unmarshal_json_response(
                models.UnprocessableResponseData, http_res
            )
            raise models.UnprocessableResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.APIError("API error occurred", http_res, http_res_text)
        if utils.match_response(http_res, "default", "application/json"):
            return models.CrmContactsDeleteResponse(
                unexpected_error_response=unmarshal_json_response(
                    Optional[models.UnexpectedErrorResponse], http_res
                ),
                http_meta=models.HTTPMetadata(request=req, response=http_res),
            )

        raise models.APIError("Unexpected response received", http_res)



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from apideck_unify.types import BaseModel
from apideck_unify.utils import FieldMetadata
from datetime import datetime
from enum import Enum
from typing import Optional
from typing_extensions import Annotated, NotRequired, TypedDict


class ExpensesFilterStatus(str, Enum):
    DRAFT = "draft"
    PENDING_APPROVAL = "pending_approval"
    APPROVED = "approved"
    POSTED = "posted"
    VOIDED = "voided"
    REJECTED = "rejected"
    DELETED = "deleted"
    OTHER = "other"


class ExpensesFilterType(str, Enum):
    EXPENSE = "expense"
    REFUND = "refund"


class ExpensesFilterTypedDict(TypedDict):
    updated_since: NotRequired[datetime]
    status: NotRequired[ExpensesFilterStatus]
    type: NotRequired[ExpensesFilterType]


class ExpensesFilter(BaseModel):
    updated_since: Annotated[Optional[datetime], FieldMetadata(query=True)] = None

    status: Annotated[Optional[ExpensesFilterStatus], FieldMetadata(query=True)] = None

    type: Annotated[Optional[ExpensesFilterType], FieldMetadata(query=True)] = None



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from apideck_unify.types import (
    BaseModel,
    Nullable,
    OptionalNullable,
    UNSET,
    UNSET_SENTINEL,
)
from apideck_unify.utils import FieldMetadata
from enum import Enum
from pydantic import model_serializer
from typing_extensions import Annotated, NotRequired, TypedDict


class InvoiceItemFilterInvoiceItemType(str, Enum):
    r"""The type of invoice item, indicating whether it is an inventory item, a service, or another type."""

    INVENTORY = "inventory"
    SERVICE = "service"
    OTHER = "other"


class InvoiceItemFilterTransactionType(str, Enum):
    r"""The kind of transaction, indicating whether it is a sales transaction or a purchase transaction."""

    SALE = "sale"
    PURCHASE = "purchase"


class InvoiceItemFilterTypedDict(TypedDict):
    type: NotRequired[Nullable[InvoiceItemFilterInvoiceItemType]]
    r"""The type of invoice item, indicating whether it is an inventory item, a service, or another type."""
    transaction_type: NotRequired[Nullable[InvoiceItemFilterTransactionType]]
    r"""The kind of transaction, indicating whether it is a sales transaction or a purchase transaction."""


class InvoiceItemFilter(BaseModel):
    type: Annotated[
        OptionalNullable[InvoiceItemFilterInvoiceItemType], FieldMetadata(query=True)
    ] = UNSET
    r"""The type of invoice item, indicating whether it is an inventory item, a service, or another type."""

    transaction_type: Annotated[
        OptionalNullable[InvoiceItemFilterTransactionType], FieldMetadata(query=True)
    ] = UNSET
    r"""The kind of transaction, indicating whether it is a sales transaction or a purchase transaction."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = ["type", "transaction_type"]
        nullable_fields = ["type", "transaction_type"]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .datascopesresources import DataScopesResources, DataScopesResourcesTypedDict
from apideck_unify.types import BaseModel
from typing_extensions import TypedDict


class ConsentRecordInputTypedDict(TypedDict):
    granted: bool
    r"""Whether consent was granted (true) or denied/revoked (false)"""
    resources: DataScopesResourcesTypedDict
    r"""Data scopes resource configuration that can be either detailed field permissions or a wildcard"""


class ConsentRecordInput(BaseModel):
    granted: bool
    r"""Whether consent was granted (true) or denied/revoked (false)"""

    resources: DataScopesResources
    r"""Data scopes resource configuration that can be either detailed field permissions or a wildcard"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .address import Address, AddressTypedDict
from .currency import Currency
from .customfield import CustomField, CustomFieldTypedDict
from .linkedcustomer import LinkedCustomer, LinkedCustomerTypedDict
from .linkedcustomer_input import LinkedCustomerInput, LinkedCustomerInputTypedDict
from .linkedtaxrate import LinkedTaxRate, LinkedTaxRateTypedDict
from .linkedtaxrate_input import LinkedTaxRateInput, LinkedTaxRateInputTypedDict
from .linkedtrackingcategory import (
    LinkedTrackingCategory,
    LinkedTrackingCategoryTypedDict,
)
from apideck_unify.types import (
    BaseModel,
    Nullable,
    OptionalNullable,
    UNSET,
    UNSET_SENTINEL,
)
from datetime import date, datetime
from enum import Enum
from pydantic import model_serializer
from typing import List, Optional
from typing_extensions import NotRequired, TypedDict


class ProjectProjectStatus(str, Enum):
    r"""Current status of the project"""

    ACTIVE = "active"
    COMPLETED = "completed"
    ON_HOLD = "on_hold"
    CANCELLED = "cancelled"
    DRAFT = "draft"
    IN_PROGRESS = "in_progress"
    APPROVED = "approved"
    OTHER = "other"


class ProjectType(str, Enum):
    r"""Type or category of the project"""

    CLIENT_PROJECT = "client_project"
    INTERNAL_PROJECT = "internal_project"
    MAINTENANCE = "maintenance"
    RESEARCH_DEVELOPMENT = "research_development"
    TRAINING = "training"
    OTHER = "other"


class Priority(str, Enum):
    r"""Priority level of the project"""

    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    CRITICAL = "critical"


class ProjectDepartmentTypedDict(TypedDict):
    r"""Department or organizational unit associated with the project"""

    id: NotRequired[str]
    r"""Unique identifier for the department"""
    name: NotRequired[str]
    r"""Name of the department"""


class ProjectDepartment(BaseModel):
    r"""Department or organizational unit associated with the project"""

    id: Optional[str] = None
    r"""Unique identifier for the department"""

    name: Optional[str] = None
    r"""Name of the department"""


class ParentProjectTypedDict(TypedDict):
    r"""Parent project if this is a subproject"""

    id: NotRequired[str]
    r"""Unique identifier for the parent project"""
    name: NotRequired[str]
    r"""Name of the parent project"""


class ParentProject(BaseModel):
    r"""Parent project if this is a subproject"""

    id: Optional[str] = None
    r"""Unique identifier for the parent project"""

    name: Optional[str] = None
    r"""Name of the parent project"""


class BillingMethod(str, Enum):
    r"""Method used for billing this project"""

    FIXED_PRICE = "fixed_price"
    TIME_AND_MATERIALS = "time_and_materials"
    MILESTONE_BASED = "milestone_based"
    RETAINER = "retainer"
    NON_BILLABLE = "non_billable"


class ProjectPhase(str, Enum):
    r"""Current phase of the project lifecycle"""

    INITIATION = "initiation"
    PLANNING = "planning"
    EXECUTION = "execution"
    MONITORING = "monitoring"
    CLOSURE = "closure"
    OTHER = "other"


class ScheduleStatus(str, Enum):
    r"""Current status of project schedule compared to plan"""

    AHEAD_OF_SCHEDULE = "ahead_of_schedule"
    ON_SCHEDULE = "on_schedule"
    BEHIND_SCHEDULE = "behind_schedule"
    CRITICAL_DELAY = "critical_delay"


class ProjectTypedDict(TypedDict):
    name: str
    r"""Name of the project"""
    id: NotRequired[str]
    r"""A unique identifier for an object."""
    downstream_id: NotRequired[Nullable[str]]
    r"""The third-party API ID of original entity"""
    display_id: NotRequired[Nullable[str]]
    r"""User-friendly project identifier"""
    reference_id: NotRequired[Nullable[str]]
    r"""External reference identifier for the project"""
    description: NotRequired[Nullable[str]]
    r"""Detailed description of the project"""
    status: NotRequired[Nullable[ProjectProjectStatus]]
    r"""Current status of the project"""
    active: NotRequired[Nullable[bool]]
    r"""Indicates whether the project is currently active or inactive"""
    project_type: NotRequired[Nullable[ProjectType]]
    r"""Type or category of the project"""
    priority: NotRequired[Nullable[Priority]]
    r"""Priority level of the project"""
    completion_percentage: NotRequired[Nullable[float]]
    r"""Percentage of project completion (0-100)"""
    start_date: NotRequired[Nullable[date]]
    r"""Start date of the project"""
    end_date: NotRequired[Nullable[date]]
    r"""Expected or actual end date of the project"""
    customer: NotRequired[Nullable[LinkedCustomerTypedDict]]
    r"""The customer this entity is linked to."""
    department: NotRequired[Nullable[ProjectDepartmentTypedDict]]
    r"""Department or organizational unit associated with the project"""
    company_id: NotRequired[Nullable[str]]
    r"""The company ID the transaction belongs to"""
    owner_id: NotRequired[Nullable[str]]
    r"""ID of the user who owns/manages this project"""
    parent_project: NotRequired[Nullable[ParentProjectTypedDict]]
    r"""Parent project if this is a subproject"""
    currency: NotRequired[Nullable[Currency]]
    r"""Indicates the associated currency for an amount of money. Values correspond to [ISO 4217](https://en.wikipedia.org/wiki/ISO_4217)."""
    budget_amount: NotRequired[Nullable[float]]
    r"""Total budgeted amount for the project"""
    approved_amount: NotRequired[Nullable[float]]
    r"""Approved budget amount for the project"""
    actual_amount: NotRequired[Nullable[float]]
    r"""Total actual amount spent on the project"""
    budget_hours: NotRequired[Nullable[float]]
    r"""Total budgeted hours for the project"""
    actual_hours: NotRequired[Nullable[float]]
    r"""Total actual hours worked on the project"""
    hourly_rate: NotRequired[Nullable[float]]
    r"""Default hourly rate for project work"""
    billing_method: NotRequired[Nullable[BillingMethod]]
    r"""Method used for billing this project"""
    is_billable: NotRequired[Nullable[bool]]
    r"""Indicates if the project is billable to the customer"""
    phase: NotRequired[Nullable[ProjectPhase]]
    r"""Current phase of the project lifecycle"""
    tax_rate: NotRequired[LinkedTaxRateTypedDict]
    tracking_categories: NotRequired[
        Nullable[List[Nullable[LinkedTrackingCategoryTypedDict]]]
    ]
    r"""A list of linked tracking categories."""
    tags: NotRequired[List[str]]
    r"""Tags associated with the project"""
    notes: NotRequired[Nullable[str]]
    r"""Additional notes about the project"""
    contract_number: NotRequired[Nullable[str]]
    r"""Contract or agreement number associated with the project"""
    profit_margin: NotRequired[Nullable[float]]
    r"""Expected profit margin percentage for the project"""
    schedule_status: NotRequired[Nullable[ScheduleStatus]]
    r"""Current status of project schedule compared to plan"""
    addresses: NotRequired[List[AddressTypedDict]]
    r"""An array of addresses associated with the project (billing, job site, etc.)"""
    team_size: NotRequired[Nullable[int]]
    r"""Number of team members assigned to the project"""
    custom_fields: NotRequired[List[CustomFieldTypedDict]]
    row_version: NotRequired[Nullable[str]]
    r"""A binary value used to detect updates to a object and prevent data conflicts. It is incremented each time an update is made to the object."""
    updated_by: NotRequired[Nullable[str]]
    r"""The user who last updated the object."""
    created_by: NotRequired[Nullable[str]]
    r"""The user who created the object."""
    created_at: NotRequired[Nullable[datetime]]
    r"""The date and time when the object was created."""
    updated_at: NotRequired[Nullable[datetime]]
    r"""The date and time when the object was last updated."""


class Project(BaseModel):
    name: str
    r"""Name of the project"""

    id: Optional[str] = None
    r"""A unique identifier for an object."""

    downstream_id: OptionalNullable[str] = UNSET
    r"""The third-party API ID of original entity"""

    display_id: OptionalNullable[str] = UNSET
    r"""User-friendly project identifier"""

    reference_id: OptionalNullable[str] = UNSET
    r"""External reference identifier for the project"""

    description: OptionalNullable[str] = UNSET
    r"""Detailed description of the project"""

    status: OptionalNullable[ProjectProjectStatus] = UNSET
    r"""Current status of the project"""

    active: OptionalNullable[bool] = UNSET
    r"""Indicates whether the project is currently active or inactive"""

    project_type: OptionalNullable[ProjectType] = UNSET
    r"""Type or category of the project"""

    priority: OptionalNullable[Priority] = UNSET
    r"""Priority level of the project"""

    completion_percentage: OptionalNullable[float] = UNSET
    r"""Percentage of project completion (0-100)"""

    start_date: OptionalNullable[date] = UNSET
    r"""Start date of the project"""

    end_date: OptionalNullable[date] = UNSET
    r"""Expected or actual end date of the project"""

    customer: OptionalNullable[LinkedCustomer] = UNSET
    r"""The customer this entity is linked to."""

    department: OptionalNullable[ProjectDepartment] = UNSET
    r"""Department or organizational unit associated with the project"""

    company_id: OptionalNullable[str] = UNSET
    r"""The company ID the transaction belongs to"""

    owner_id: OptionalNullable[str] = UNSET
    r"""ID of the user who owns/manages this project"""

    parent_project: OptionalNullable[ParentProject] = UNSET
    r"""Parent project if this is a subproject"""

    currency: OptionalNullable[Currency] = UNSET
    r"""Indicates the associated currency for an amount of money. Values correspond to [ISO 4217](https://en.wikipedia.org/wiki/ISO_4217)."""

    budget_amount: OptionalNullable[float] = UNSET
    r"""Total budgeted amount for the project"""

    approved_amount: OptionalNullable[float] = UNSET
    r"""Approved budget amount for the project"""

    actual_amount: OptionalNullable[float] = UNSET
    r"""Total actual amount spent on the project"""

    budget_hours: OptionalNullable[float] = UNSET
    r"""Total budgeted hours for the project"""

    actual_hours: OptionalNullable[float] = UNSET
    r"""Total actual hours worked on the project"""

    hourly_rate: OptionalNullable[float] = UNSET
    r"""Default hourly rate for project work"""

    billing_method: OptionalNullable[BillingMethod] = UNSET
    r"""Method used for billing this project"""

    is_billable: OptionalNullable[bool] = True
    r"""Indicates if the project is billable to the customer"""

    phase: OptionalNullable[ProjectPhase] = UNSET
    r"""Current phase of the project lifecycle"""

    tax_rate: Optional[LinkedTaxRate] = None

    tracking_categories: OptionalNullable[List[Nullable[LinkedTrackingCategory]]] = (
        UNSET
    )
    r"""A list of linked tracking categories."""

    tags: Optional[List[str]] = None
    r"""Tags associated with the project"""

    notes: OptionalNullable[str] = UNSET
    r"""Additional notes about the project"""

    contract_number: OptionalNullable[str] = UNSET
    r"""Contract or agreement number associated with the project"""

    profit_margin: OptionalNullable[float] = UNSET
    r"""Expected profit margin percentage for the project"""

    schedule_status: OptionalNullable[ScheduleStatus] = UNSET
    r"""Current status of project schedule compared to plan"""

    addresses: Optional[List[Address]] = None
    r"""An array of addresses associated with the project (billing, job site, etc.)"""

    team_size: OptionalNullable[int] = UNSET
    r"""Number of team members assigned to the project"""

    custom_fields: Optional[List[CustomField]] = None

    row_version: OptionalNullable[str] = UNSET
    r"""A binary value used to detect updates to a object and prevent data conflicts. It is incremented each time an update is made to the object."""

    updated_by: OptionalNullable[str] = UNSET
    r"""The user who last updated the object."""

    created_by: OptionalNullable[str] = UNSET
    r"""The user who created the object."""

    created_at: OptionalNullable[datetime] = UNSET
    r"""The date and time when the object was created."""

    updated_at: OptionalNullable[datetime] = UNSET
    r"""The date and time when the object was last updated."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = [
            "id",
            "downstream_id",
            "display_id",
            "reference_id",
            "description",
            "status",
            "active",
            "project_type",
            "priority",
            "completion_percentage",
            "start_date",
            "end_date",
            "customer",
            "department",
            "company_id",
            "owner_id",
            "parent_project",
            "currency",
            "budget_amount",
            "approved_amount",
            "actual_amount",
            "budget_hours",
            "actual_hours",
            "hourly_rate",
            "billing_method",
            "is_billable",
            "phase",
            "tax_rate",
            "tracking_categories",
            "tags",
            "notes",
            "contract_number",
            "profit_margin",
            "schedule_status",
            "addresses",
            "team_size",
            "custom_fields",
            "row_version",
            "updated_by",
            "created_by",
            "created_at",
            "updated_at",
        ]
        nullable_fields = [
            "downstream_id",
            "display_id",
            "reference_id",
            "description",
            "status",
            "active",
            "project_type",
            "priority",
            "completion_percentage",
            "start_date",
            "end_date",
            "customer",
            "department",
            "company_id",
            "owner_id",
            "parent_project",
            "currency",
            "budget_amount",
            "approved_amount",
            "actual_amount",
            "budget_hours",
            "actual_hours",
            "hourly_rate",
            "billing_method",
            "is_billable",
            "phase",
            "tracking_categories",
            "notes",
            "contract_number",
            "profit_margin",
            "schedule_status",
            "team_size",
            "row_version",
            "updated_by",
            "created_by",
            "created_at",
            "updated_at",
        ]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m


class ProjectInputTypedDict(TypedDict):
    name: str
    r"""Name of the project"""
    display_id: NotRequired[Nullable[str]]
    r"""User-friendly project identifier"""
    reference_id: NotRequired[Nullable[str]]
    r"""External reference identifier for the project"""
    description: NotRequired[Nullable[str]]
    r"""Detailed description of the project"""
    status: NotRequired[Nullable[ProjectProjectStatus]]
    r"""Current status of the project"""
    active: NotRequired[Nullable[bool]]
    r"""Indicates whether the project is currently active or inactive"""
    project_type: NotRequired[Nullable[ProjectType]]
    r"""Type or category of the project"""
    priority: NotRequired[Nullable[Priority]]
    r"""Priority level of the project"""
    completion_percentage: NotRequired[Nullable[float]]
    r"""Percentage of project completion (0-100)"""
    start_date: NotRequired[Nullable[date]]
    r"""Start date of the project"""
    end_date: NotRequired[Nullable[date]]
    r"""Expected or actual end date of the project"""
    customer: NotRequired[Nullable[LinkedCustomerInputTypedDict]]
    r"""The customer this entity is linked to."""
    department: NotRequired[Nullable[ProjectDepartmentTypedDict]]
    r"""Department or organizational unit associated with the project"""
    company_id: NotRequired[Nullable[str]]
    r"""The company ID the transaction belongs to"""
    owner_id: NotRequired[Nullable[str]]
    r"""ID of the user who owns/manages this project"""
    parent_project: NotRequired[Nullable[ParentProjectTypedDict]]
    r"""Parent project if this is a subproject"""
    currency: NotRequired[Nullable[Currency]]
    r"""Indicates the associated currency for an amount of money. Values correspond to [ISO 4217](https://en.wikipedia.org/wiki/ISO_4217)."""
    budget_amount: NotRequired[Nullable[float]]
    r"""Total budgeted amount for the project"""
    approved_amount: NotRequired[Nullable[float]]
    r"""Approved budget amount for the project"""
    budget_hours: NotRequired[Nullable[float]]
    r"""Total budgeted hours for the project"""
    hourly_rate: NotRequired[Nullable[float]]
    r"""Default hourly rate for project work"""
    billing_method: NotRequired[Nullable[BillingMethod]]
    r"""Method used for billing this project"""
    is_billable: NotRequired[Nullable[bool]]
    r"""Indicates if the project is billable to the customer"""
    phase: NotRequired[Nullable[ProjectPhase]]
    r"""Current phase of the project lifecycle"""
    tax_rate: NotRequired[LinkedTaxRateInputTypedDict]
    tracking_categories: NotRequired[
        Nullable[List[Nullable[LinkedTrackingCategoryTypedDict]]]
    ]
    r"""A list of linked tracking categories."""
    tags: NotRequired[List[str]]
    r"""Tags associated with the project"""
    notes: NotRequired[Nullable[str]]
    r"""Additional notes about the project"""
    contract_number: NotRequired[Nullable[str]]
    r"""Contract or agreement number associated with the project"""
    profit_margin: NotRequired[Nullable[float]]
    r"""Expected profit margin percentage for the project"""
    schedule_status: NotRequired[Nullable[ScheduleStatus]]
    r"""Current status of project schedule compared to plan"""
    addresses: NotRequired[List[AddressTypedDict]]
    r"""An array of addresses associated with the project (billing, job site, etc.)"""
    team_size: NotRequired[Nullable[int]]
    r"""Number of team members assigned to the project"""
    custom_fields: NotRequired[List[CustomFieldTypedDict]]
    row_version: NotRequired[Nullable[str]]
    r"""A binary value used to detect updates to a object and prevent data conflicts. It is incremented each time an update is made to the object."""


class ProjectInput(BaseModel):
    name: str
    r"""Name of the project"""

    display_id: OptionalNullable[str] = UNSET
    r"""User-friendly project identifier"""

    reference_id: OptionalNullable[str] = UNSET
    r"""External reference identifier for the project"""

    description: OptionalNullable[str] = UNSET
    r"""Detailed description of the project"""

    status: OptionalNullable[ProjectProjectStatus] = UNSET
    r"""Current status of the project"""

    active: OptionalNullable[bool] = UNSET
    r"""Indicates whether the project is currently active or inactive"""

    project_type: OptionalNullable[ProjectType] = UNSET
    r"""Type or category of the project"""

    priority: OptionalNullable[Priority] = UNSET
    r"""Priority level of the project"""

    completion_percentage: OptionalNullable[float] = UNSET
    r"""Percentage of project completion (0-100)"""

    start_date: OptionalNullable[date] = UNSET
    r"""Start date of the project"""

    end_date: OptionalNullable[date] = UNSET
    r"""Expected or actual end date of the project"""

    customer: OptionalNullable[LinkedCustomerInput] = UNSET
    r"""The customer this entity is linked to."""

    department: OptionalNullable[ProjectDepartment] = UNSET
    r"""Department or organizational unit associated with the project"""

    company_id: OptionalNullable[str] = UNSET
    r"""The company ID the transaction belongs to"""

    owner_id: OptionalNullable[str] = UNSET
    r"""ID of the user who owns/manages this project"""

    parent_project: OptionalNullable[ParentProject] = UNSET
    r"""Parent project if this is a subproject"""

    currency: OptionalNullable[Currency] = UNSET
    r"""Indicates the associated currency for an amount of money. Values correspond to [ISO 4217](https://en.wikipedia.org/wiki/ISO_4217)."""

    budget_amount: OptionalNullable[float] = UNSET
    r"""Total budgeted amount for the project"""

    approved_amount: OptionalNullable[float] = UNSET
    r"""Approved budget amount for the project"""

    budget_hours: OptionalNullable[float] = UNSET
    r"""Total budgeted hours for the project"""

    hourly_rate: OptionalNullable[float] = UNSET
    r"""Default hourly rate for project work"""

    billing_method: OptionalNullable[BillingMethod] = UNSET
    r"""Method used for billing this project"""

    is_billable: OptionalNullable[bool] = True
    r"""Indicates if the project is billable to the customer"""

    phase: OptionalNullable[ProjectPhase] = UNSET
    r"""Current phase of the project lifecycle"""

    tax_rate: Optional[LinkedTaxRateInput] = None

    tracking_categories: OptionalNullable[List[Nullable[LinkedTrackingCategory]]] = (
        UNSET
    )
    r"""A list of linked tracking categories."""

    tags: Optional[List[str]] = None
    r"""Tags associated with the project"""

    notes: OptionalNullable[str] = UNSET
    r"""Additional notes about the project"""

    contract_number: OptionalNullable[str] = UNSET
    r"""Contract or agreement number associated with the project"""

    profit_margin: OptionalNullable[float] = UNSET
    r"""Expected profit margin percentage for the project"""

    schedule_status: OptionalNullable[ScheduleStatus] = UNSET
    r"""Current status of project schedule compared to plan"""

    addresses: Optional[List[Address]] = None
    r"""An array of addresses associated with the project (billing, job site, etc.)"""

    team_size: OptionalNullable[int] = UNSET
    r"""Number of team members assigned to the project"""

    custom_fields: Optional[List[CustomField]] = None

    row_version: OptionalNullable[str] = UNSET
    r"""A binary value used to detect updates to a object and prevent data conflicts. It is incremented each time an update is made to the object."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = [
            "display_id",
            "reference_id",
            "description",
            "status",
            "active",
            "project_type",
            "priority",
            "completion_percentage",
            "start_date",
            "end_date",
            "customer",
            "department",
            "company_id",
            "owner_id",
            "parent_project",
            "currency",
            "budget_amount",
            "approved_amount",
            "budget_hours",
            "hourly_rate",
            "billing_method",
            "is_billable",
            "phase",
            "tax_rate",
            "tracking_categories",
            "tags",
            "notes",
            "contract_number",
            "profit_margin",
            "schedule_status",
            "addresses",
            "team_size",
            "custom_fields",
            "row_version",
        ]
        nullable_fields = [
            "display_id",
            "reference_id",
            "description",
            "status",
            "active",
            "project_type",
            "priority",
            "completion_percentage",
            "start_date",
            "end_date",
            "customer",
            "department",
            "company_id",
            "owner_id",
            "parent_project",
            "currency",
            "budget_amount",
            "approved_amount",
            "budget_hours",
            "hourly_rate",
            "billing_method",
            "is_billable",
            "phase",
            "tracking_categories",
            "notes",
            "contract_number",
            "profit_margin",
            "schedule_status",
            "team_size",
            "row_version",
        ]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .basesdk import BaseSDK
from .sdkconfiguration import SDKConfiguration
from apideck_unify.apideck_companies import ApideckCompanies
from apideck_unify.apideck_departments import ApideckDepartments
from apideck_unify.employeepayrolls import EmployeePayrolls
from apideck_unify.employees import Employees
from apideck_unify.employeeschedules_sdk import EmployeeSchedulesSDK
from apideck_unify.payrolls import Payrolls
from apideck_unify.timeoffrequests import TimeOffRequests
from typing import Optional


class Hris(BaseSDK):
    employees: Employees
    companies: ApideckCompanies
    departments: ApideckDepartments
    payrolls: Payrolls
    employee_payrolls: EmployeePayrolls
    employee_schedules: EmployeeSchedulesSDK
    time_off_requests: TimeOffRequests

    def __init__(
        self, sdk_config: SDKConfiguration, parent_ref: Optional[object] = None
    ) -> None:
        BaseSDK.__init__(self, sdk_config, parent_ref=parent_ref)
        self.sdk_configuration = sdk_config
        self._init_sdks()

    def _init_sdks(self):
        self.employees = Employees(self.sdk_configuration, parent_ref=self.parent_ref)
        self.companies = ApideckCompanies(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.departments = ApideckDepartments(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.payrolls = Payrolls(self.sdk_configuration, parent_ref=self.parent_ref)
        self.employee_payrolls = EmployeePayrolls(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.employee_schedules = EmployeeSchedulesSDK(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.time_off_requests = TimeOffRequests(
            self.sdk_configuration, parent_ref=self.parent_ref
        )



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from enum import Enum


class ConsentState(str, Enum):
    r"""The current consent state of the connection"""

    IMPLICIT = "implicit"
    PENDING = "pending"
    GRANTED = "granted"
    DENIED = "denied"
    REVOKED = "revoked"
    REQUIRES_RECONSENT = "requires_reconsent"



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .basesdk import BaseSDK
from .sdkconfiguration import SDKConfiguration
from apideck_unify.agedcreditors_sdk import AgedCreditorsSDK
from apideck_unify.ageddebtors_sdk import AgedDebtorsSDK
from apideck_unify.attachments import Attachments
from apideck_unify.balancesheet_sdk import BalanceSheetSDK
from apideck_unify.bankaccounts import BankAccounts
from apideck_unify.bankfeedaccounts import BankFeedAccounts
from apideck_unify.bankfeedstatements import BankFeedStatements
from apideck_unify.billpayments import BillPayments
from apideck_unify.bills import Bills
from apideck_unify.categories import Categories
from apideck_unify.companyinfo_sdk import CompanyInfoSDK
from apideck_unify.creditnotes import CreditNotes
from apideck_unify.customers import Customers
from apideck_unify.departments import Departments
from apideck_unify.expenses import Expenses
from apideck_unify.invoiceitems import InvoiceItems
from apideck_unify.invoices import Invoices
from apideck_unify.journalentries import JournalEntries
from apideck_unify.ledgeraccounts import LedgerAccounts
from apideck_unify.locations import Locations
from apideck_unify.payments import Payments
from apideck_unify.profitandloss_sdk import ProfitAndLossSDK
from apideck_unify.projects import Projects
from apideck_unify.purchaseorders import PurchaseOrders
from apideck_unify.quotes import Quotes
from apideck_unify.subsidiaries import Subsidiaries
from apideck_unify.suppliers import Suppliers
from apideck_unify.taxrates import TaxRates
from apideck_unify.trackingcategories import TrackingCategories
from typing import Optional


class Accounting(BaseSDK):
    tax_rates: TaxRates
    bills: Bills
    invoices: Invoices
    ledger_accounts: LedgerAccounts
    invoice_items: InvoiceItems
    credit_notes: CreditNotes
    customers: Customers
    suppliers: Suppliers
    payments: Payments
    company_info: CompanyInfoSDK
    balance_sheet: BalanceSheetSDK
    profit_and_loss: ProfitAndLossSDK
    journal_entries: JournalEntries
    purchase_orders: PurchaseOrders
    subsidiaries: Subsidiaries
    locations: Locations
    departments: Departments
    attachments: Attachments
    bank_accounts: BankAccounts
    tracking_categories: TrackingCategories
    bill_payments: BillPayments
    expenses: Expenses
    aged_creditors: AgedCreditorsSDK
    aged_debtors: AgedDebtorsSDK
    bank_feed_accounts: BankFeedAccounts
    bank_feed_statements: BankFeedStatements
    categories: Categories
    quotes: Quotes
    projects: Projects

    def __init__(
        self, sdk_config: SDKConfiguration, parent_ref: Optional[object] = None
    ) -> None:
        BaseSDK.__init__(self, sdk_config, parent_ref=parent_ref)
        self.sdk_configuration = sdk_config
        self._init_sdks()

    def _init_sdks(self):
        self.tax_rates = TaxRates(self.sdk_configuration, parent_ref=self.parent_ref)
        self.bills = Bills(self.sdk_configuration, parent_ref=self.parent_ref)
        self.invoices = Invoices(self.sdk_configuration, parent_ref=self.parent_ref)
        self.ledger_accounts = LedgerAccounts(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.invoice_items = InvoiceItems(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.credit_notes = CreditNotes(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.customers = Customers(self.sdk_configuration, parent_ref=self.parent_ref)
        self.suppliers = Suppliers(self.sdk_configuration, parent_ref=self.parent_ref)
        self.payments = Payments(self.sdk_configuration, parent_ref=self.parent_ref)
        self.company_info = CompanyInfoSDK(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.balance_sheet = BalanceSheetSDK(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.profit_and_loss = ProfitAndLossSDK(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.journal_entries = JournalEntries(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.purchase_orders = PurchaseOrders(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.subsidiaries = Subsidiaries(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.locations = Locations(self.sdk_configuration, parent_ref=self.parent_ref)
        self.departments = Departments(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.attachments = Attachments(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.bank_accounts = BankAccounts(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.tracking_categories = TrackingCategories(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.bill_payments = BillPayments(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.expenses = Expenses(self.sdk_configuration, parent_ref=self.parent_ref)
        self.aged_creditors = AgedCreditorsSDK(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.aged_debtors = AgedDebtorsSDK(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.bank_feed_accounts = BankFeedAccounts(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.bank_feed_statements = BankFeedStatements(
            self.sdk_configuration, parent_ref=self.parent_ref
        )
        self.categories = Categories(self.sdk_configuration, parent_ref=self.parent_ref)
        self.quotes = Quotes(self.sdk_configuration, parent_ref=self.parent_ref)
        self.projects = Projects(self.sdk_configuration, parent_ref=self.parent_ref)



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .httpmetadata import HTTPMetadata, HTTPMetadataTypedDict
from .unexpectederrorresponse import (
    UnexpectedErrorResponse,
    UnexpectedErrorResponseTypedDict,
)
from .updateuploadsessionresponse import (
    UpdateUploadSessionResponse,
    UpdateUploadSessionResponseTypedDict,
)
from apideck_unify.types import BaseModel
from apideck_unify.utils import (
    FieldMetadata,
    HeaderMetadata,
    PathParamMetadata,
    QueryParamMetadata,
    RequestMetadata,
)
import io
import pydantic
from typing import IO, Optional, Union
from typing_extensions import Annotated, NotRequired, TypedDict


FILE_STORAGE_UPLOAD_SESSIONS_UPLOAD_OP_SERVERS = [
    "https://upload.apideck.com",
]


class FileStorageUploadSessionsUploadGlobalsTypedDict(TypedDict):
    consumer_id: NotRequired[str]
    r"""ID of the consumer which you want to get or push data from"""
    app_id: NotRequired[str]
    r"""The ID of your Unify application"""


class FileStorageUploadSessionsUploadGlobals(BaseModel):
    consumer_id: Annotated[
        Optional[str],
        pydantic.Field(alias="x-apideck-consumer-id"),
        FieldMetadata(header=HeaderMetadata(style="simple", explode=False)),
    ] = None
    r"""ID of the consumer which you want to get or push data from"""

    app_id: Annotated[
        Optional[str],
        pydantic.Field(alias="x-apideck-app-id"),
        FieldMetadata(header=HeaderMetadata(style="simple", explode=False)),
    ] = None
    r"""The ID of your Unify application"""


class FileStorageUploadSessionsUploadRequestTypedDict(TypedDict):
    id: str
    r"""ID of the record you are acting upon."""
    part_number: float
    r"""Part number of the file part being uploaded."""
    request_body: Union[bytes, IO[bytes], io.BufferedReader]
    consumer_id: NotRequired[str]
    r"""ID of the consumer which you want to get or push data from"""
    app_id: NotRequired[str]
    r"""The ID of your Unify application"""
    service_id: NotRequired[str]
    r"""Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API."""
    digest: NotRequired[str]
    r"""The RFC3230 message digest of the uploaded part. Only required for the Box connector. More information on the Box API docs [here](https://developer.box.com/reference/put-files-upload-sessions-id/#param-digest)"""
    raw: NotRequired[bool]
    r"""Include raw response. Mostly used for debugging purposes"""


class FileStorageUploadSessionsUploadRequest(BaseModel):
    id: Annotated[
        str, FieldMetadata(path=PathParamMetadata(style="simple", explode=False))
    ]
    r"""ID of the record you are acting upon."""

    part_number: Annotated[
        float, FieldMetadata(query=QueryParamMetadata(style="form", explode=True))
    ]
    r"""Part number of the file part being uploaded."""

    request_body: Annotated[
        Union[bytes, IO[bytes], io.BufferedReader],
        FieldMetadata(request=RequestMetadata(media_type="*/*")),
    ]

    consumer_id: Annotated[
        Optional[str],
        pydantic.Field(alias="x-apideck-consumer-id"),
        FieldMetadata(header=HeaderMetadata(style="simple", explode=False)),
    ] = None
    r"""ID of the consumer which you want to get or push data from"""

    app_id: Annotated[
        Optional[str],
        pydantic.Field(alias="x-apideck-app-id"),
        FieldMetadata(header=HeaderMetadata(style="simple", explode=False)),
    ] = None
    r"""The ID of your Unify application"""

    service_id: Annotated[
        Optional[str],
        pydantic.Field(alias="x-apideck-service-id"),
        FieldMetadata(header=HeaderMetadata(style="simple", explode=False)),
    ] = None
    r"""Provide the service id you want to call (e.g., pipedrive). Only needed when a consumer has activated multiple integrations for a Unified API."""

    digest: Annotated[
        Optional[str],
        FieldMetadata(header=HeaderMetadata(style="simple", explode=False)),
    ] = None
    r"""The RFC3230 message digest of the uploaded part. Only required for the Box connector. More information on the Box API docs [here](https://developer.box.com/reference/put-files-upload-sessions-id/#param-digest)"""

    raw: Annotated[
        Optional[bool],
        FieldMetadata(query=QueryParamMetadata(style="form", explode=True)),
    ] = False
    r"""Include raw response. Mostly used for debugging purposes"""


class FileStorageUploadSessionsUploadResponseTypedDict(TypedDict):
    http_meta: HTTPMetadataTypedDict
    update_upload_session_response: NotRequired[UpdateUploadSessionResponseTypedDict]
    r"""UploadSessions"""
    unexpected_error_response: NotRequired[UnexpectedErrorResponseTypedDict]
    r"""Unexpected error"""


class FileStorageUploadSessionsUploadResponse(BaseModel):
    http_meta: Annotated[Optional[HTTPMetadata], pydantic.Field(exclude=True)] = None

    update_upload_session_response: Optional[UpdateUploadSessionResponse] = None
    r"""UploadSessions"""

    unexpected_error_response: Optional[UnexpectedErrorResponse] = None
    r"""Unexpected error"""
